/** 
 * ContainerPrj.cpp 
 * 
 * Copyright (c) 2006, Amanjit Singh Gill
 * All rights reserved.
 * 
 * Redistribution and use in source and binary  forms, with or without
 * modification, are permitted provided  that the following conditions
 * are met:
 *
 * Redistributions of  source  code must  retain  the  above copyright
 * notice, this  list  of conditions   and the  following  disclaimer.
 * Redistributions in  binary form must  reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or   other     materials   provided  with     the
 * distribution. Neither the name of Amanjtit Singh Gill nor the names
 * of its contributors   may be used  to  endorse or promote  products
 * derived      from this software      without specific prior written
 * permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS"  AND ANY EXPRESS OR  IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO,  THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS
 * FOR  A PARTICULAR PURPOSE  ARE DISCLAIMED.  IN  NO  EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL,   SPECIAL,   EXEMPLARY,     OR CONSEQUENTIAL    DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT  OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR  PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT  LIABILITY,  OR TORT   (INCLUDING  NEGLIGENCE OR  OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifdef _WIN32 
# undef DO_LEAKCHECK // win32 vld 
# ifdef DO_LEAKCHECK
#  include "vld.h"
# endif
#endif

#include "HashMap.h"
#include "Version.h"
#include <sstream>
#include <map>
#include "math.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#ifdef _WIN32
# include <windows.h>
# include <tchar.h>
# include <malloc.h>
#else
# include <sys/param.h>  
# include <sys/times.h>
# include <sys/time.h>
# include <sys/types.h>
#endif

using namespace ag;

/*******************************************************************/

#ifdef _WIN32
/** Win32 Timer impl */
struct Timer {
  long currTick() {
    return ::GetTickCount();
  } 
};

#else
/** POSIX timer impl */
struct Timer {
  long currTick() {
    struct timeval tv;
    gettimeofday( &tv, NULL );
    return (tv.tv_sec * 1000 + tv.tv_usec / 1000);
  }
};

#endif

/**
 * My Stopwatch
 */
template<class TIMER_>
class StopWatch {
public:
  StopWatch() {  
    m_lTick = m_timer.currTick();
  }  
  long stop() {
    return m_timer.currTick() - m_lTick;
  }
  void start() {
    m_lTick = m_timer.currTick();    
  }
private:
  TIMER_ m_timer;
  long   m_lTick;  
};


typedef StopWatch<Timer> StopWatch_t;

#define CLOCKSMSEC   (CLOCKS_PER_SEC / 1000.0F)
#define GETMSEC(f,s) ((double)(f - s) / CLOCKSMSEC)

const std::string g_sLine = "----------------";

/*******************************************************************/

static void DispHash(const std::string& sVal) {
  std::cout << std::endl << sVal << " " << (unsigned long) g_strHashFunc(sVal); 
}

static char getRandomChar() {
  int iRandom = rand();
  int iRest   = iRandom % 2;  
  if (!iRest) 
    return (char) (iRandom % (122-65)) + 65;  
  return (char) (iRandom % (57-48)) + 48;    
}

static void getRandomString(std::string& sOutString) {
  sOutString = "";
  for (int x=0; x<12;x++)
    sOutString += getRandomChar();
}

/*******************************************************************/

/**
 * Main method 
 * @todo: rather messy, turn into a unit test
 */
#ifdef _WIN32
  int _tmain()
#else
  int main() 
#endif
{    
  #ifdef _WIN32
    _set_sbh_threshold(128);   
    ULONG HeapFragValue = 2;
    HeapSetInformation((HANDLE)_get_heap_handle(), 
      HeapCompatibilityInformation, 
      &HeapFragValue, 
      sizeof(HeapFragValue));      

//    int tmpFlag = _CrtSetDbgFlag( _CRTDBG_REPORT_FLAG ); 
//      tmpFlag |=  _CRTDBG_CHECK_ALWAYS_DF; // check heap on every alloc
//    _CrtSetDbgFlag( tmpFlag );
   
  #endif  
  StopWatch_t timer;
  std::cout << std::endl << "Container version: " << CONTAINER_VER;
  std::cout << std::endl << "built: " << (__DATE__) << " / " << (__TIME__);
  std::cout << std::endl; std::cout.flush();  

  const int iRUNS = 1000000; // you may want to adjust this   

  std::string sRand = "";
  std::vector<std::string> lstStrings;
  std::vector<std::string> lstNameStrings;
  getRandomString(sRand);
  lstStrings.reserve(iRUNS);
  lstNameStrings.reserve(iRUNS);
  std::cout << std::endl << sRand << std::endl; std::cout.flush();  
  std::cout << std::endl << "Generating Random Names "; std::cout.flush();
  for (int z=0; z<iRUNS; z++) {
    getRandomString(sRand);
    lstStrings.push_back(sRand);
  }
  std::cout << std::endl << "Generating Predef Names "; std::cout.flush();
  char cbBuffer[4096];  
  for (int z=0; z<iRUNS; z++) {
    sprintf(&cbBuffer[0], "%d Name %d", z, z);     
    int iDiv = iRUNS/20;
    if (iDiv > 0) {
      if ((z % iDiv) == 0) {
        std::cout << "."; std::cout.flush();
      }
    }
    lstNameStrings.push_back(&cbBuffer[0]);
  }
  std::cout << " done. " << std::endl; std::cout.flush();      

  hash_map<std::string, long, StrHashFunc> myMap_tst; // /*iRUNS*/);       
  typedef hash_map<std::string, long, StrHashFunc>::iterator       iter_t;
  typedef hash_map<std::string, long, StrHashFunc>::const_iterator const_iter_t;

  std::string sName = ("22");  
  std::string xName = ("22");    

  std::string sOne("One"); 
  std::string sTwo("Two"); 
  DispHash(sOne);
  DispHash(sTwo);
  DispHash("0");
  DispHash("Onko");
  DispHash("oknO");  
  DispHash("ONKO");
  DispHash("OKNO");  
  DispHash("Name 0");  
  DispHash("Name 1");  
  DispHash("Name 12");  
  DispHash("Name 112");  
  DispHash("Name 305");  
  std::cout.flush();

  myMap_tst[sOne] = 1;
  std::cout << std::endl << "---------" << std::flush;
  myMap_tst.dumpHash_();  
  myMap_tst[sTwo] = 2;  
  std::cout << std::endl << "---------" << std::flush;
  myMap_tst.dumpHash_();  

  iter_t pSearch;
  pSearch = myMap_tst.find(sOne);
  assert(pSearch != myMap_tst.end());
  assert(myMap_tst.find(sTwo) != myMap_tst.end());  
  myMap_tst[sOne] = 1;
  std::cout << std::endl << "---------" << std::flush;
  myMap_tst.dumpHash_();  
  iter_t pIt = myMap_tst.find(sOne);
  assert(pIt->second == 1);
  pIt = myMap_tst.find(sTwo);
  assert(pIt->second == 2);

  assert(myMap_tst.find("Zonk") == myMap_tst.end());

  myMap_tst.insert(make_pair("Punkt", 111));
  assert(myMap_tst.find("Punkt") != myMap_tst.end());
  pIt = myMap_tst.find("Punkt");
  assert(pIt->second == 111);
  assert(myMap_tst.hasValue(111));
  myMap_tst.erase(myMap_tst.find("Punkt"));
  assert(myMap_tst.find("Punkt") == myMap_tst.end());
  assert(!myMap_tst.hasValue(111));  // 111 should not be used by someone else

  pIt = myMap_tst.find(sOne);
  assert(pIt->second == 1); 

  // check iteration
  for (int x=0; x<10; x++) {
    myMap_tst.insert(make_pair(lstNameStrings[x], x));
  }
  
  std::cout << std::endl << g_sLine << " Dumped Hash " << std::endl; std::cout.flush();
  myMap_tst.dumpHash_();  
  std::cout << std::endl << g_sLine << " Dumped Hash via iteration" << std::endl; std::cout.flush();

  hash_map<std::string, long, StrHashFunc>::iterator pBegin, pEnd;
  for (pBegin = myMap_tst.begin(), pEnd = myMap_tst.end();
       pBegin != pEnd;
      ++pBegin) {
      std::cout << std::endl << "{" << pBegin->first << "} --> {" << pBegin->second << "}"; 
  }
  
  std::cout << std::endl << g_sLine << " Creating copy of Hash and comparing" << std::endl; std::cout.flush();
  hash_map<std::string, long, StrHashFunc> mapCopy(myMap_tst);
  hash_map<std::string, long, StrHashFunc>::iterator pBegin2, pEnd2;
  for (pBegin = myMap_tst.begin(), pEnd = myMap_tst.end(), pBegin2 = mapCopy.begin(), pEnd2 = mapCopy.end();
       pBegin != pEnd && pBegin2 != pEnd2;
       ++pBegin, ++pBegin2) {
      assert(pBegin->first  == pBegin2->first);
      assert(pBegin->second == pBegin2->second); 
  }
  assert(pBegin == pEnd);
  assert(pBegin2 == pEnd2);
  assert(mapCopy == myMap_tst); // == comparision
  
  std::cout << std::endl << " ... orig ... ";
  myMap_tst.dumpHash_();
  mapCopy = myMap_tst;          // assignment operator+#
  std::cout << std::endl << " ... orig after copy ... ";
  myMap_tst.dumpHash_();
  std::cout << std::endl << " ... vs ... ";
  mapCopy.dumpHash_();  
  assert(mapCopy == myMap_tst);
  myMap_tst.clear();
  assert(myMap_tst.getCount() == 0);
  std::cout << std::endl << " ... ";
  mapCopy.dumpHash_();
  mapCopy.clear();  
    
  mapCopy.insert(make_pair("ForCopy2A", 111));
  mapCopy.insert(make_pair("ForCopy2B", 222));
  mapCopy.dumpHash_();
  hash_map<std::string, long, StrHashFunc> mapCopy2;
  swap(mapCopy, mapCopy2);
  assert(mapCopy.size() == 0);  
  pIt = mapCopy2.find("ForCopy2A");
  assert(pIt->second == 111);
  pIt = mapCopy2.find("ForCopy2B");
  assert(pIt->second == 222);
  
  // make_pair.
  std::pair<hash_map<std::string, long, StrHashFunc>::iterator, bool> bRet;
  bRet = mapCopy2.insert(make_pair("APair", 666));
  bRet = mapCopy2.insert(make_pair("APair2", 222));
  assert(bRet.second == false);
  pIt = mapCopy2.find("APair");
  assert(pIt->second==666);
  bRet = mapCopy2.insert(make_pair("APair", 666));
  assert(bRet.second == true);
  pIt = mapCopy2.find("APair");
  assert(pIt->second ==666);
  assert(((bRet.first)->first)  == "APair");
  assert(((bRet.first)->second) == 666);

  // container.find
  hash_map<std::string, long, StrHashFunc>::iterator pItFind = mapCopy2.find("APair2");
  assert(pItFind != mapCopy2.end());
  assert(pItFind->first == "APair2");
  assert(pItFind->second == 222);
  
  // BUGGY 
#if 0 // const_iterator does not work yet
  hash_map<std::string, long, StrHashFunc>::const_iterator pItFind2 = mapCopy2.find("APair");
  assert(pItFind->first == "APair");
  assert(pItFind->second == 666);   
#endif

  for (int r=0;r<2;r++) {
    hash_map<std::string, long, StrHashFunc>  myMap; // /*iRUNS*/);       
    if (r % 2) {
      std::cout << std::endl << std::endl << g_sLine << g_sLine << " again, setting size this time... " << std::endl; std::cout.flush();            
      myMap.growCapacity(iRUNS);
    }    

    // v1
    std::cout << std::endl << g_sLine << "Putting " << std::endl; std::cout.flush();
    timer.start();
    for (int x=0; x< iRUNS;x++) {    
      myMap.insert(make_pair(lstNameStrings[x], x));
    }  
    double dDur = timer.stop();    
    std::cout << std::endl << "Time was: " << dDur << " for " << iRUNS << " Elements.";            
   
    std::map<std::string, long> stlMap ;
    timer.start(); 
    for (int x=0; x<iRUNS;x++) {    
      stlMap[lstNameStrings[x]] = x;
    }
    dDur = timer.stop();    
    std::cout << std::endl << "Time for stlmap was: " << dDur << " for " << iRUNS << " Elements.";        

    std::cout << std::endl << g_sLine << " Getting " << std::endl; std::cout.flush();  
    pIt = myMap.find("2 Name 2");     
    assert(pIt != myMap.end());
    assert(pIt->second == 2);
    pIt = myMap.find("7 Name 7");
    assert(pIt != myMap.end());
    assert(pIt->second == 7); 

    timer.start();   
    for (int x=0; x< iRUNS;x++) {    
      pIt = myMap.find(lstNameStrings[x]);
      assert(pIt->second == x);    
    } 
    dDur = timer.stop();
    std::cout << std::endl << "Time was: " << dDur << " for " << iRUNS << " Elements.";    

    timer.start(); 
    for (int x=0; x< iRUNS;x++) {          
      long lVal = -1;
      lVal = stlMap[lstNameStrings[x]];
      assert(lVal == x);          
    } 
    dDur = timer.stop();
    std::cout << std::endl << "Time for stlmap was: " << dDur << " for " << iRUNS << " Elements.";

    // std::cout << std::endl << "We had " << myMap.getCollisions() << " Collisions"; 
  }

  std::cout << std::endl << std::endl << std::endl;

  for (int r=0;r<2;r++) {
    std::map<std::string, long> stlMap2 ;
    hash_map<std::string, long, StrHashFunc>  myMap2; // (/*iRUNS*/);      
    if (r) {
      std::cout << std::endl << std::endl << g_sLine << g_sLine << " again, setting size this time... " << std::endl; std::cout.flush();
      myMap2.growCapacity(iRUNS);      
    }
    // again for our randomized strings
    std::cout << std::endl << g_sLine << " Putting randstrings: " << std::endl; std::cout.flush();
    timer.start(); 
    for (int x=0; x< iRUNS;x++) {    
      myMap2.insert(make_pair(lstStrings[x], x));
    }  
    double dDur = timer.stop();
    std::cout << std::endl << "randstrings: Time was: " << dDur << " for " << iRUNS << " Elements.";    
    // myMap2.dumpHash_();
    assert(myMap2.getCount() == (unsigned long) iRUNS);    

    timer.start(); 
    for (int x=0; x<iRUNS;x++) {    
      stlMap2[lstStrings[x]] = x;
    }
    dDur = timer.stop();

    std::cout << std::endl << "randstrings: Time for stlMap2 was: " << dDur << " for " << iRUNS << " Elements." << std::endl;       

    std::cout << std::endl << g_sLine << " Getting randstrings: " << std::endl; std::cout.flush();          
    timer.start(); 
    for (int x=0; x< iRUNS;x++) {          
      pIt = myMap2.find(lstStrings[x]);
      assert(pIt->second == x);    
    } 
    dDur = timer.stop();
    std::cout << std::endl << "randstrings: Time was: " << dDur << " for " << iRUNS << " Elements.";    

    long lVal;
    timer.start(); 
    for (int x=0; x< iRUNS;x++) {                
      lVal = stlMap2[lstStrings[x]];
      assert(lVal == x);    
    } 
    dDur = timer.stop();
    std::cout << std::endl << "randstrings: Time for stlMap2 was: " << dDur << " for " << iRUNS << " Elements.";          
  }      

  hash_map<std::string, std::string, StrHashFunc> mapNames;
  typedef hash_map<std::string, std::string, StrHashFunc>::iterator iter2_t;
  mapNames.insert(make_pair("peter","lustig"));
  mapNames.insert(make_pair("rainer","zufall"));
  mapNames.insert(make_pair("dr.","mabuse"));
  mapNames.dumpHash_();
  
  iter2_t pIt2 = mapNames.find("peter");
  assert(pIt2->second == "lustig");

  std::cout << std::endl << "done"; std::cout.flush();  

  int i; std::cin >> i;
  return 0;	
}

