
/*******************************************************************/

const int    iHASH_INITSIZE = 12;    /**< initial hash size                */
const double dGROWTH_FACT   = 1.25;  /**< growth factor of hashmap buckets */

namespace ag {

/**
 * A Generic hash_map implementation
 * @author Amanjit Singh Gill (C)2006
 * More or less close to the TR-1. NO, THIS IS NOT
 * PRODUCTION READY. This is just my shot at writing
 * a TR-1 "alike" hashmap. Love it or leave it.
 */
template <class KEY_,
  class VAL_,
  class HASHFUNC_,
  class EQUAL_ = std::equal_to<KEY_>,
  class ALLOC_ = CustomAlloc< std::pair<KEY_,VAL_> > >
class hash_map {
public:
  /** default constructor */
  hash_map() :
      m_ulCount(0), m_ulCapacity(0), m_rgBuckets(), m_hashFunc(), m_eqFunc() {
    initBuckets_(iHASH_INITSIZE);
  }

  /** init constructor with specific size */
  hash_map(unsigned long ulSize) :
      m_ulCount(0), m_ulCapacity(0), m_rgBuckets(), m_hashFunc(), m_eqFunc() {
    initBuckets_(ulSize);
  }

  /** copy constructor */
  hash_map(const hash_map& srcMap) :
      m_ulCount(0), m_ulCapacity(0), m_rgBuckets(), m_hashFunc(), m_eqFunc() {
    initBuckets_(srcMap.getCapacity());
    typename hash_map<KEY_, VAL_, HASHFUNC_, EQUAL_, ALLOC_>::const_iterator pCurr, pEnd;
    for (pCurr = srcMap.begin(), pEnd = srcMap.end();
         pCurr != pEnd;
         ++pCurr) {
      insert(*pCurr);
    }
  }

  /** assignment operator */
  hash_map& operator=(const hash_map& srcMap) {
    if (&srcMap == this)
      return *this;
    clear();
    initBuckets_(srcMap.getCapacity());
    typename hash_map<KEY_, VAL_, HASHFUNC_, EQUAL_, ALLOC_>::const_iterator pCurr, pEnd;
    for (pCurr = srcMap.begin(), pEnd = srcMap.end();
         pCurr != pEnd;
         ++pCurr) {
      insert(*pCurr);
    }
    return *this;
  }

  /** destructor */
  virtual ~hash_map() {
  }

  /**
   * Public typedefs, see the TR-1 unordered_map spec for
   * more info.
   */
  typedef ALLOC_                           allocator_type;
  typedef typename ALLOC_::value_type      value_type;
  typedef typename ALLOC_::reference       reference;
  typedef typename ALLOC_::const_reference const_reference;
  typedef typename ALLOC_::pointer         pointer;
  typedef typename ALLOC_::const_pointer   const_pointer;
  typedef typename ALLOC_::size_type       size_type;
  typedef typename ALLOC_::difference_type difference_type;


  typedef KEY_                             key_type;
  typedef VAL_                             data_type;
  typedef HASHFUNC_                        hasher;
  typedef EQUAL_                           key_equal;

  /** helper: invoke the hash function */
  unsigned long doHashFunc(const KEY_& key) const {
    return m_hashFunc(key);
  }

  /** set an entry via the [] operaor */
  inline VAL_& operator[](const KEY_& key) {
    unsigned long ulHash = doHashFunc(key);
    if (++m_ulCount*dGROWTH_FACT > m_ulCapacity)
      initBuckets_((unsigned long)(m_ulCount*dGROWTH_FACT*1.25));
    unsigned long lIter = ulHash % m_ulCapacity;
    typename std::vector<Bucket, CustomAlloc<Bucket> >::iterator pCurr = (m_rgBuckets.begin() + lIter);
    typename std::list< value_type, ALLOC_ >::iterator pSecIt, pSecEnd;
    for (pSecIt = pCurr->m_lstSecBucks.begin(), pSecEnd = pCurr->m_lstSecBucks.end();
        pSecIt != pSecEnd;
        ++pSecIt) {
      if (m_eqFunc(pSecIt->first,key)) { // overwriting existing key
        m_ulCount--;
        pSecIt->second = VAL_();
        return pSecIt->second;
      }
    }
    pCurr->m_lstSecBucks.push_back(make_pair(key, VAL_()));
    pSecIt = pCurr->m_lstSecBucks.end();
    pSecIt--;
    return pSecIt->second;
  }

  /** whether a key exists */
  inline bool hasKey_(const KEY_& key) const {
    unsigned long ulHash = doHashFunc(key);
    unsigned long lIter  = ulHash % m_ulCapacity;
    typename std::vector<Bucket, CustomAlloc<Bucket> >::const_iterator pCurr = (m_rgBuckets.begin() + lIter);
    typename std::list< value_type, ALLOC_ >::const_iterator pSecIt, pSecEnd;
    for (pSecIt = pCurr->m_lstSecBucks.begin(), pSecEnd = pCurr->m_lstSecBucks.end();
        pSecIt != pSecEnd;
        ++pSecIt) {
      if (m_eqFunc(pSecIt->first,key))
        return true;
    }
    return false;
  }

  /** whether a value exists */
  inline bool hasValue(const VAL_& val) const {
    typename std::vector<Bucket, CustomAlloc<Bucket> >::const_iterator pCurr = (m_rgBuckets.begin());
    size_t lRuns = m_ulCount;
    while(pCurr != m_rgBuckets.end() && lRuns > 0) {
        for (typename std::list< value_type, ALLOC_ >::const_iterator pSecIt = pCurr->m_lstSecBucks.begin();
          pSecIt != pCurr->m_lstSecBucks.end();
          ++pSecIt) {
        if (pSecIt->second == val)
          return true;
      }
      lRuns -= pCurr->m_lstSecBucks.size();
      pCurr++;
    }
    return false;
  }

  /** remove a key */
  inline bool delKey_(const KEY_& key) {
    unsigned long ulHash = doHashFunc(key);
    unsigned long lIter  = ulHash % m_ulCapacity;
    typename std::vector<Bucket, CustomAlloc<Bucket> >::iterator pCurr = (m_rgBuckets.begin() + lIter);
     for (typename std::list< value_type, ALLOC_ >::iterator pSecIt = pCurr->m_lstSecBucks.begin();
          pSecIt != pCurr->m_lstSecBucks.end();
          ++pSecIt) {
      if (m_eqFunc(pSecIt->first,key)) {
        pCurr->m_lstSecBucks.erase(pSecIt);
        m_ulCount--;
        return true;
      }
    }
    return false;
  }

  /** retrieve amount of entries */
  inline unsigned long getCount() const {
    return m_ulCount;
  }

  /** set the capacity of our map */
  inline void setCapacity(unsigned long ulSize) {
    if (ulSize == 0) {
      m_rgBuckets.resize(0);
      m_ulCapacity = 0;
      m_ulCount    = 0;
      return;
    }
    if (ulSize*dGROWTH_FACT > m_ulCapacity)
      initBuckets_((unsigned long)(ulSize*dGROWTH_FACT*1.25));
  }

  /** retrieve the capacity of our map */
  inline unsigned long getCapacity() const {
    return m_ulCapacity;
  }

  /** helper: dump the hash map */
  void dumpHash_() const {
    std::cout << std::endl << "Dumping hash... " << std::endl;
    typename std::list< value_type, ALLOC_ >::const_iterator pSecIt, pSecEnd;
    int iIter = 0;
    for (unsigned int x=0; x<m_ulCapacity;++x) {
      for (pSecIt = m_rgBuckets[x].m_lstSecBucks.begin(), pSecEnd = m_rgBuckets[x].m_lstSecBucks.end();
           pSecIt != pSecEnd;
           ++pSecIt) {
          std::cout << x << ": {\""
          << pSecIt->first  << "\" => \""
          << pSecIt->second << "\"} " << std::endl;
          iIter++;
      }
    }
    std::cout.flush();
    assert(iIter == int(m_ulCount));
  }

  /** init our buckets */
  inline void initBuckets_(unsigned long lEstimSize) {
    m_ulCapacity = getNextPrime(lEstimSize);
    m_rgBuckets.resize(m_ulCapacity);
    assert(m_rgBuckets.capacity() >= m_ulCapacity);
    reHash_();
  }

  /** actually put an entry */
  inline void putPrivate_(const std::pair<KEY_,VAL_>& entry) {
    unsigned long ulHash = doHashFunc(entry.first);
    unsigned long lIter  = ulHash % m_ulCapacity;
    typename std::list< value_type, ALLOC_ >::iterator pSecIt, pSecEnd;
    typename std::vector<Bucket, CustomAlloc<Bucket> >::iterator pCurr = (m_rgBuckets.begin() + lIter);
    for (pSecIt = pCurr->m_lstSecBucks.begin(), pSecEnd = pCurr->m_lstSecBucks.end();
        pSecIt != pSecEnd;
        ++pSecIt) {
      if (m_eqFunc(pSecIt->first,entry.first)) {
          pSecIt->second = entry.second;
        return;
      }
    }
    pCurr->m_lstSecBucks.push_back(entry);
  }

  /** rehash our map */
  inline void reHash_() {
    typename std::vector<Bucket, CustomAlloc<Bucket> > tmpBuckets(m_rgBuckets);
    m_rgBuckets.resize(0);
    m_rgBuckets.resize(m_ulCapacity);
    typename std::list< value_type, ALLOC_ >::const_iterator pSecIt, pSecEnd;    
    for (unsigned int x=0; x<m_ulCapacity;x++) {
      for (pSecIt = tmpBuckets[x].m_lstSecBucks.begin(), pSecEnd = tmpBuckets[x].m_lstSecBucks.end();
          pSecIt != pSecEnd;
          ++pSecIt) {
        putPrivate_(*pSecIt);        
      }
    }
  }

  /** clear the hashmap */
  inline void clear() {
    setCapacity(0);
    m_ulCount = 0;
  }

  /** resize the hashmap */
  void resize(size_type n) {
    setCapacity(n);
  }

  /** return the hasher */
  hasher hash_funct() const {
    return m_hashFunc;
  }

  /** return the key comparison function */
  key_equal key_eq() const {
    return m_eqFunc;
  }

  /** return amount of buckets */
  inline size_type bucket_count() const {
    return size();
  }

  /** return the capacity */
  inline size_type max_size() const {
    return m_ulCapacity;
  }

  /** return amount of entries */
  inline size_type size() const {
    return m_ulCount;
  }

  /** is the map empty ? */
  inline bool empty() const {
      return m_ulCount == 0 ? true : false;
  }

  /** return amount of entries for a key */
  inline size_type count(const KEY_& key) const {
    return hasKey_(key) ? 1 : 0;
  }

  struct SecondBuck;

  /**
   * Bucket entry
   */
  struct Bucket {
    Bucket()
      : m_lstSecBucks() {
    }

    std::list< value_type, ALLOC_ > m_lstSecBucks;
  };

  unsigned long       m_ulCount;                         /**< amount of real entries in map */
  unsigned long       m_ulCapacity;                      /**< capacity = dim of buckets */
  std::vector<Bucket, CustomAlloc<Bucket> > m_rgBuckets; /**< dim is always m_ulCapacity */
  HASHFUNC_           m_hashFunc;                        /**< the hash function */
  EQUAL_              m_eqFunc;                          /**< the comparison function */

  /******************************************************************/

  /**
   * Base iterator template class
   */
  template < bool ISCONST_ >
  class IteratorBase  {
  public:
    // why everybody loves templates... ;)
    /** typedef for our bucket list */
    typedef typename std::vector<
      typename hash_map<KEY_,VAL_,HASHFUNC_,EQUAL_,ALLOC_>::Bucket,
                CustomAlloc< typename hash_map<KEY_,VAL_,HASHFUNC_,EQUAL_,ALLOC_>::Bucket > >
                                                bucklist_t;
    /** typedef for our secondary bucket */
    typedef typename std::list<
      typename hash_map<KEY_,VAL_,HASHFUNC_,EQUAL_,ALLOC_>::value_type, ALLOC_ >
                                                secbuck_t;

    typedef forward_iterator_tag  iterator_category;
    typedef typename ALLOC_::value_type         value_type;
    typedef typename ALLOC_::difference_type    difference_type;
    typedef typename MySelect< ISCONST_,  // Select pointer or const_pointer to T.
      const_pointer,
      typename ALLOC_::pointer >::result        pointer;
    typedef typename MySelect< ISCONST_,  // Select reference or const_reference to T.
      const_reference,
      typename ALLOC_::reference >::result      reference;

  private:
    bucklist_t*                                 m_pLstMyBucks;
    typename bucklist_t::iterator               m_pBuck;
    typename secbuck_t::iterator                m_pEntry;

  public:
    /** default constructor */
    IteratorBase() : 
      m_pLstMyBucks(0), m_pBuck(), m_pEntry() {
    }
    
    /** construct from list and position (begin/end) */
    IteratorBase(bucklist_t* pList, bool bEnd) :
      m_pLstMyBucks(pList), m_pBuck(m_pLstMyBucks->end()), m_pEntry() {
      if (!bEnd) {
        if (!m_pLstMyBucks->empty()) {
          m_pBuck  = m_pLstMyBucks->begin();
          m_pEntry = m_pBuck->m_lstSecBucks.begin();
        }
      }
    }

    /** construct a specific SecBuck position. the caller must assure that the list is not empty */
    IteratorBase(bucklist_t* pList, typename bucklist_t::iterator pItBuckPos,
        typename secbuck_t::iterator pItSecBuckPos) :
      m_pLstMyBucks(pList), m_pBuck(pItBuckPos), m_pEntry(pItSecBuckPos) {
    }

    /** return designated value. */
    reference	operator*() const {
      return (*m_pEntry);
    }

    /** Return pointer to class object, delegate to operator '*' */
    pointer operator->() const {
      return &(**this);	//
    }

    /** increment operator */
    IteratorBase& operator++() {
      m_pEntry++;
      if (m_pEntry == m_pBuck->m_lstSecBucks.end())
        checkpEntryNextBuck_();
      return *this;
    }

    /** increment const operator */
    const IteratorBase operator++(int iNum) {
      for (int x=0; x<iNum; x++)
        this++;	        // Increment
      return m_pEntry;	// Return copy of original
    }

    //
    // TODO cleanup operators
    //

    bool operator ==(IteratorBase<ISCONST_>& itR) const {
      assert(itR.m_pLstMyBucks->size() != 0);
      assert(m_pLstMyBucks->size() != 0);
      if (itR.m_pBuck == itR.m_pLstMyBucks->end()) {
        return ( (m_pBuck != m_pLstMyBucks->end()) ? false : true );
      }
      if (m_pBuck == m_pLstMyBucks->end()) {
        return ( (itR.m_pBuck != itR.m_pLstMyBucks->end()) ? false : true );
      }
      // normal case, no end() situation
      return ( (m_pEntry == itR.m_pEntry) &&
              (m_pBuck == itR.m_pBuck) );
    }

    bool operator ==(const IteratorBase<ISCONST_>& itR) const { // const
      assert(itR.m_pLstMyBucks->size() != 0);
      assert(m_pLstMyBucks->size() != 0);
      if (itR.m_pBuck == itR.m_pLstMyBucks->end()) {
        return ( (m_pBuck != m_pLstMyBucks->end()) ? false : true );
      }
      if (m_pBuck == m_pLstMyBucks->end()) {
        return ( (itR.m_pBuck != itR.m_pLstMyBucks->end()) ? false : true );
      }
      // normal case, no end() situation
      return ( (m_pEntry == itR.m_pEntry) &&
              (m_pBuck == itR.m_pBuck) );
    }

    bool operator !=(IteratorBase<ISCONST_>& itR) const {  // non-const
      assert(itR.m_pLstMyBucks->size() != 0);
      assert(m_pLstMyBucks->size() != 0);
      if (itR.m_pBuck == itR.m_pLstMyBucks->end()) {
        return ( (m_pBuck == m_pLstMyBucks->end()) ? false : true );
      }
      if (m_pBuck == m_pLstMyBucks->end()) {
        return ( (itR.m_pBuck == itR.m_pLstMyBucks->end()) ? false : true );
      }
      // normal case, no end() situation
      return ( (m_pEntry == itR.m_pEntry) &&
              (m_pBuck == itR.m_pBuck) );
    }

    bool operator !=(const IteratorBase<ISCONST_>& itR) const {  // non-const
      assert(itR.m_pLstMyBucks->size() != 0);
      assert(m_pLstMyBucks->size() != 0);
      if (itR.m_pBuck == itR.m_pLstMyBucks->end()) {
        return ( (m_pBuck == m_pLstMyBucks->end()) ? false : true );
      }
      if (m_pBuck == m_pLstMyBucks->end()) {
        return ( (itR.m_pBuck == itR.m_pLstMyBucks->end()) ? false : true );
      }
      // normal case, no end() situation
      return ( (m_pEntry == itR.m_pEntry) &&
              (m_pBuck == itR.m_pBuck) );
    }

    /** check if our secbucks iterator is at a sane position */
    void checkpEntryNextBuck_() {
      if (m_pEntry != m_pBuck->m_lstSecBucks.end())
        return;
      m_pBuck++;
      while (m_pBuck != m_pLstMyBucks->end())
        if (m_pBuck->m_lstSecBucks.empty())
          m_pBuck++;
        else
          break;
      if (m_pBuck != m_pLstMyBucks->end())
        m_pEntry = m_pBuck->m_lstSecBucks.begin();
    }

    #ifdef _WIN32
      friend class IteratorBase; // gcc 3.x doesn't really like this
    #endif
    friend class hash_map;
  };

  /******************************************************************/

  typedef IteratorBase< false > iterator;       /**< the normal mutable iterator */
  typedef IteratorBase< true >  const_iterator; /**< the const operator */

  iterator begin() {
    if (m_ulCount == 0)
      return end();
    return IteratorBase<false>(&m_rgBuckets, false);
  }

	const_iterator begin() const {
    if (m_ulCount == 0)
      return end();
    return IteratorBase<true>(
      const_cast< std::vector<Bucket, CustomAlloc<Bucket> >* >(&m_rgBuckets),
      false);
  }

	iterator end() {
    return IteratorBase<false>(
      const_cast< std::vector<Bucket, CustomAlloc<Bucket> >* >(&m_rgBuckets),
      true);
  }

	const_iterator end() const {
    return IteratorBase<true>(
      const_cast< std::vector<Bucket, CustomAlloc<Bucket> >* >(&m_rgBuckets),
      true);
  }

  iterator find(const KEY_& key) {
    unsigned long ulHash = doHashFunc(key);
    unsigned long lIter  = ulHash % m_ulCapacity;
    typename std::vector<Bucket, CustomAlloc<Bucket> >::iterator pCurr = (m_rgBuckets.begin() + lIter);
    typename std::list< value_type, ALLOC_ >::iterator pSecIt, pSecEnd;
    for (pSecIt = pCurr->m_lstSecBucks.begin(), pSecEnd = pCurr->m_lstSecBucks.end();
         pSecIt != pSecEnd;
        ++pSecIt) {
      if (m_eqFunc(pSecIt->first,key)) {
        return iterator(&m_rgBuckets, pCurr, pSecIt);
      }
    }
    return end();
  }

  const_iterator find(const KEY_& key) const {
    unsigned long ulHash = doHashFunc(key);
    unsigned long lIter  = ulHash % m_ulCapacity;
    typename std::vector<Bucket, CustomAlloc<Bucket> >::const_iterator pCurr = (m_rgBuckets.begin() + lIter);
    typename std::list< value_type, ALLOC_ >::const_iterator pSecIt, pSecEnd;
    for (pSecIt = pCurr->m_lstSecBucks.begin(), pSecEnd = pCurr->m_lstSecBucks.end();
         pSecIt != pSecEnd;
        ++pSecIt) {
      if (m_eqFunc(pSecIt->first,key)) {
        const_iterator itRet;
        itRet.m_pBuck      = pCurr;
        itRet.m_pBuckEnd   = m_rgBuckets.end();
        itRet.m_pBuckBegin = m_rgBuckets.begin();
        itRet.m_pEntry     = pSecIt;
        return itRet;
      }
    }
    return end();
  }

  inline pair<iterator, bool> insert(const value_type& x) {
    pair<iterator, bool> pairRet;
    unsigned long ulHash = doHashFunc(x.first);
    if (++m_ulCount*dGROWTH_FACT > m_ulCapacity)
      initBuckets_((unsigned long)(m_ulCount*dGROWTH_FACT*1.25));
    unsigned long lIter = ulHash % m_ulCapacity;
    typename std::vector<Bucket, CustomAlloc<Bucket> >::iterator pCurr = (m_rgBuckets.begin() + lIter);
    typename std::list< value_type, ALLOC_ >::iterator pSecIt, pSecEnd;
    for (pSecIt = pCurr->m_lstSecBucks.begin(), pSecEnd = pCurr->m_lstSecBucks.end();
        pSecIt != pSecEnd;
        ++pSecIt) {
      if (m_eqFunc(pSecIt->first,x.first)) { // overwriting existing key
        m_ulCount--;
        pSecIt->second = x.second;
        pairRet.second = true;
        pairRet.first  = IteratorBase<false>(&m_rgBuckets, pCurr, pSecIt);
        return pairRet;
      }
    }
    pCurr->m_lstSecBucks.push_back(x);
    pairRet.second = false;
    pairRet.first  = IteratorBase<false>(&m_rgBuckets, pCurr, pCurr->m_lstSecBucks.end()--);
    return pairRet;
  }

  void erase(iterator pos) {
    delKey_(pos->first);
  }

};